package cn.fw.morax.server.task;

import cn.fw.common.cache.locker.DistributedLocker;
import cn.fw.morax.common.constant.Constant;
import cn.fw.morax.common.constant.TimeTaskConstant;
import cn.fw.morax.common.utils.PublicUtil;
import cn.fw.morax.domain.db.eval.*;
import cn.fw.morax.domain.db.kpi.IndicatorUserValue;
import cn.fw.morax.domain.enums.*;
import cn.fw.morax.service.biz.CommonService;
import cn.fw.morax.service.data.eval.*;
import cn.fw.morax.service.data.kpi.IndicatorUserValueService;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.google.common.collect.Maps;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.common.filter.impl.Op;
import org.redisson.api.RLock;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.TransactionStatus;
import org.springframework.transaction.annotation.Transactional;

import java.math.BigDecimal;
import java.time.LocalDate;
import java.time.YearMonth;
import java.time.temporal.TemporalAdjusters;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Lock;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * @author : jiangchao
 * @className : KpiReportRatioTask
 * @description : 绩效报表定时器
 * @date : 2022-04-07 15:29
 */
@Component
@Slf4j
@RequiredArgsConstructor
@ConditionalOnProperty(prefix = "task", name = "switch", havingValue = "on")
public class EvalUserReportTask {

    private final EvalGroupIndicatorTargetHitLogService evalGroupIndicatorTargetHitLogService;
    private final EvalGroupIndicatorHitLogService evalGroupIndicatorHitLogService;
    private final EvalGroupIndicatorParamService evalGroupIndicatorParamService;
    private final EvalUserRankCodeReportService evalUserRankCodeReportService;
    private final EvalUserRankStageReportService evalUserRankStageReportService;
    private final EvalUserValueReportService evalUserValueReportService;
    private final EvalGroupIndicatorService evalGroupIndicatorService;
    private final IndicatorUserValueService indicatorUserValueService;
    private final EvalIndicatorValueService evalIndicatorValueService;
    private final PlatformTransactionManager platformTransactionManager;
    private final EvalUserRankReportService evalUserRankReportService;
    private final EvalGroupRankStageService evalGroupRankStageService;
    private final EvalUserReportService evalUserReportService;
    private final TransactionDefinition transactionDefinition;
    private final EvalUserPoolService evalUserPoolService;
    private final EvalShopPoolService evalShopPoolService;
    private final EvalGroupRankService evalGroupRankService;
    private final EvalGroupService evalGroupService;
    private final DistributedLocker distributedLocker;
    private final CommonService commonService;

    @Value("${spring.cache.custom.global-prefix}:eval:user:report")
    @Getter
    private String evalUserReportDistKey;

    /**
     * 考评报表定时任务
     *
     */
    @Scheduled(cron = TimeTaskConstant.EVAL_REPORT)
    @Transactional(rollbackFor = Exception.class)
    public void evalUserReportTask() {
//        this.evalUserReport(LocalDate.now().minusDays(1));
    }

    /**
     * 绩效报表定时任务
     * 1. 遍历考评排名组
     * 2. 创建考评排名组报表
     * 3. 创建考评排名组编码
     * 4. 统计数据
     *
     * @param date
     */
    public void evalUserReport(LocalDate date) {
        Lock lock = distributedLocker.lock(getEvalUserReportDistKey());
        if (! ((RLock) lock).isLocked()) {
            return;
        }
        TransactionStatus transactionStatus = platformTransactionManager.getTransaction(transactionDefinition);
        try {
            log.info("定时任务【考评人员报表数据抽取】开始执行");
            Map<Long, List<EvalGroupRank>> groupEvalGroupMap = queryEffectRank(date);
            for (Map.Entry<Long, List<EvalGroupRank>> entry : groupEvalGroupMap.entrySet()) {
                for (EvalGroupRank rank : entry.getValue()) {
                    //清除当月数据（当月只有一条数据）
                    cleanRankCurMonthData(rank.getId(), date);
                    //抽取排名组数据
                    createRankReport(rank, entry.getKey(), date);
                }
            }
            platformTransactionManager.commit(transactionStatus);
        } catch (Exception e){
            platformTransactionManager.rollback(transactionStatus);
            log.error("考评人员报表数据抽取失败， " + e.getMessage(), e);
        } finally {
            lock.unlock();
        }
    }

    /**
     * 计算排名组报表数据
     * @param rank
     * @param groupId
     * @param date
     */
    public void createRankReport(EvalGroupRank rank, Long groupId, LocalDate date) {
        //查已经开始的考评阶段
        List<EvalGroupRankStage> stages = evalGroupRankStageService.getRankStartStages(rank.getId(), date);
        if (PublicUtil.isEmpty(stages)) {
            return;
        }
        //排名组
        EvalUserRankReport reportRank = this.saveRankReport(rank, date, groupId);
        final Long evalUserRankReportId = reportRank.getId();
        for (EvalGroupRankStage rankStage : stages) {
            List<EvalGroup> evalGroups = evalGroupService.getEvalGroups(rank.getId());
            List<Long> evalGroupIds = evalGroups.stream().map(EvalGroup::getId).collect(Collectors.toList());
            if (PublicUtil.isEmpty(evalGroupIds)) {
                continue;
            }
            List<EvalUserPool> userPools = evalUserPoolService.list(Wrappers.<EvalUserPool>lambdaQuery()
                    .in(EvalUserPool::getEvalGroupId, evalGroupIds)
                    .eq(EvalUserPool::getMonthly, YearMonth.from(date))
                    .eq(EvalUserPool::getYn, Boolean.TRUE)
            );
            List<EvalShopPool> shopPools = evalShopPoolService.list(Wrappers.<EvalShopPool>lambdaQuery()
                    .in(EvalShopPool::getEvalGroupId, evalGroupIds)
                    .eq(EvalShopPool::getMonthly, YearMonth.from(date))
                    .eq(EvalShopPool::getYn, Boolean.TRUE)
            );
            List<EvalGroupIndicator> evalGroupIndicators = evalGroupIndicatorService.list(Wrappers.<EvalGroupIndicator>lambdaQuery()
                    .in(EvalGroupIndicator::getEvalGroupId, evalGroupIds)
                    .eq(EvalGroupIndicator::getYn, Boolean.TRUE)
                    .orderByAsc(EvalGroupIndicator::getId)
            );

            EvalUserRankStageReport rankStageReport = this.saveRankStageReport(reportRank, rank, rankStage, evalGroups);
            final Long evalUserRankStageReportId = rankStageReport.getId();
            //排名组指标编码、人员、门店
            List<EvalUserRankCodeReport> rankCodes = this.saveRankCodeReports(evalGroupIndicators, evalUserRankReportId, evalUserRankStageReportId, date);
            List<EvalUserReport> evalUserReports = this.saveStaffReports(userPools, date, evalUserRankReportId, evalUserRankStageReportId);
            List<EvalUserReport> evalShopReports = this.saveShopReports(shopPools, date, evalUserRankReportId, evalUserRankStageReportId);
            //人员维度
            List<EvalUserValueReport> reportUserValues = calcUser(userPools, evalGroupIndicators, evalUserReports, date);
            //门店维度
            calcShop(shopPools, evalGroupIndicators, evalShopReports, date);
        }


    }
    public List<EvalUserValueReport> calcUser(List<EvalUserPool> pools,
                                              List<EvalGroupIndicator> evalGroupIndicators,
                                              List<EvalUserReport> evalUserReports,
                                              LocalDate date) {
        List<EvalUserValueReport> reportUserValues = new ArrayList<>();
        List<Long> userIds = pools.stream().map(EvalUserPool::getUserId).distinct().collect(Collectors.toList());
        List<Long> poolIds = pools.stream().map(EvalUserPool::getId).collect(Collectors.toList());
        List<Long> evalGroupIndicatorIds = evalGroupIndicators.stream().map(EvalGroupIndicator::getId).collect(Collectors.toList());
        List<EvalGroupIndicatorHitLog> hitLogs = evalGroupIndicatorHitLogService.getHitLogs(evalGroupIndicatorIds, poolIds, EvalScopeEnum.STAFF, date);
        Map<Long, List<EvalGroupIndicatorHitLog>> indicatorHitLogMap = hitLogs.stream().collect(Collectors.groupingBy(EvalGroupIndicatorHitLog::getEvalGroupIndicatorId));
        Map<Long, Long> userPoolReportIdMap = evalUserReports.stream().collect(Collectors.toMap(EvalUserReport::getEvalUserPoolId, EvalUserReport::getId, (v1, v2) -> v1));

        for (EvalGroupIndicator groupIndicator : evalGroupIndicators) {
            final ScoreWayEnum scoreWay = groupIndicator.getScoreWay();
            final IndicatorCodeTypeEnum codeType = groupIndicator.getCodeType();
            final Long evalGroupIndicatorId = groupIndicator.getId();
            List<EvalGroupIndicatorParam> params = evalGroupIndicatorParamService.list(Wrappers.<EvalGroupIndicatorParam>lambdaQuery()
                    .eq(EvalGroupIndicatorParam::getEvalGroupIndicatorId, evalGroupIndicatorId)
                    .eq(EvalGroupIndicatorParam::getYn, Boolean.TRUE)
            );

            List<EvalGroupIndicatorHitLog> userHitLogs = indicatorHitLogMap.getOrDefault(evalGroupIndicatorId, new ArrayList<>());
            Map<Long, EvalGroupIndicatorHitLog> poolHitLogMap = userHitLogs.stream()
                    .collect(Collectors.toMap(EvalGroupIndicatorHitLog::getPoolId, Function.identity(), (v1, v2) -> v1));
            for (EvalUserPool pool : pools) {
                if (! pool.getEvalGroupId().equals(groupIndicator.getEvalGroupId())) {
                    continue;
                }

                Optional<EvalGroupIndicatorHitLog> hitLog = Optional.ofNullable(poolHitLogMap.get(pool.getId()));
                if (pool.getDataDate().compareTo(date) != 0) {
                    hitLog = evalGroupIndicatorHitLogService.getHitLog(evalGroupIndicatorId, pool.getId(), EvalScopeEnum.STAFF, pool.getDataDate());
                }
                BigDecimal value = (ScoreWayEnum.LADDER.equals(scoreWay)) ?
                        hitLog.map(EvalGroupIndicatorHitLog::getHitLadderValue).orElse(BigDecimal.ZERO) :
                        hitLog.map(EvalGroupIndicatorHitLog::getHitCommissionValue).orElse(BigDecimal.ZERO);
                Long evalUserReportId = userPoolReportIdMap.get(pool.getId());
                reportUserValues.add(convertDB(evalUserReportId, groupIndicator, value, date, params));
            }

            //是否有子指标
            if (IndicatorCodeTypeEnum.COMBINE_INDICATOR.equals(codeType)) {
                for (EvalGroupIndicatorParam param : params) {
                    Map<Long, EvalGroupIndicatorTargetHitLog> targetHitLogMap = queryTargetValue(param.getId(), poolIds, date, EvalScopeEnum.STAFF);
                    for (EvalUserPool pool : pools) {
                        Boolean staffChange = pool.getDataDate().compareTo(date) != 0;
                        EvalGroupIndicatorTargetHitLog targetHitLog = (staffChange)
                                ? queryTargetValue(param.getId(), pool.getId(), pool.getDataDate(), EvalScopeEnum.STAFF) : targetHitLogMap.get(pool.getId());
                        BigDecimal value = this.getParamValue(targetHitLog);
                        Long evalUserReportId = userPoolReportIdMap.get(pool.getId());
                        reportUserValues.add(convertParamDB(evalUserReportId, param, value, date));
                    }
                }
            }
        }
        if (PublicUtil.isNotEmpty(reportUserValues)) {
            evalUserValueReportService.saveBatch(reportUserValues);
        }
        return reportUserValues;
    }

    public List<EvalUserValueReport> calcShop(List<EvalShopPool> pools,
                                              List<EvalGroupIndicator> evalGroupIndicators,
                                              List<EvalUserReport> evalShopReports,
                                              LocalDate date) {
        List<EvalUserValueReport> reportUserValues = new ArrayList<>();
        List<Long> shopIds = pools.stream().map(EvalShopPool::getShopId).distinct().collect(Collectors.toList());
        List<Long> poolIds = pools.stream().map(EvalShopPool::getId).collect(Collectors.toList());
        List<Long> evalGroupIndicatorIds = evalGroupIndicators.stream().map(EvalGroupIndicator::getId).collect(Collectors.toList());
        List<EvalGroupIndicatorHitLog> hitLogs = evalGroupIndicatorHitLogService.getHitLogs(evalGroupIndicatorIds, poolIds, EvalScopeEnum.SHOP, date);
        Map<Long, List<EvalGroupIndicatorHitLog>> indicatorHitLogMap = hitLogs.stream().collect(Collectors.groupingBy(EvalGroupIndicatorHitLog::getEvalGroupIndicatorId));
        Map<Long, Long> shopPoolReportIdMap = evalShopReports.stream().collect(Collectors.toMap(EvalUserReport::getEvalShopPoolId, EvalUserReport::getId, (v1, v2) -> v1));

        for (EvalGroupIndicator groupIndicator : evalGroupIndicators) {
            final ScoreWayEnum scoreWay = groupIndicator.getScoreWay();
            final IndicatorCodeTypeEnum codeType = groupIndicator.getCodeType();
            final Long evalGroupIndicatorId = groupIndicator.getId();
            List<EvalGroupIndicatorParam> params = evalGroupIndicatorParamService.list(Wrappers.<EvalGroupIndicatorParam>lambdaQuery()
                    .eq(EvalGroupIndicatorParam::getEvalGroupIndicatorId, evalGroupIndicatorId)
                    .eq(EvalGroupIndicatorParam::getYn, Boolean.TRUE)
            );

            List<EvalGroupIndicatorHitLog> userHitLogs = indicatorHitLogMap.getOrDefault(evalGroupIndicatorId, new ArrayList<>());
            Map<Long, EvalGroupIndicatorHitLog> poolHitLogMap = userHitLogs.stream()
                    .collect(Collectors.toMap(EvalGroupIndicatorHitLog::getPoolId, Function.identity(), (v1, v2) -> v1));
            for (EvalShopPool pool : pools) {
                if (! pool.getEvalGroupId().equals(groupIndicator.getEvalGroupId())) {
                    continue;
                }

                Optional<EvalGroupIndicatorHitLog> hitLog = Optional.ofNullable(poolHitLogMap.get(pool.getId()));
                if (pool.getDataDate().compareTo(date) != 0) {
                    hitLog = evalGroupIndicatorHitLogService.getHitLog(evalGroupIndicatorId, pool.getId(), EvalScopeEnum.SHOP, pool.getDataDate());
                }
                BigDecimal value = (ScoreWayEnum.LADDER.equals(scoreWay)) ?
                        hitLog.map(EvalGroupIndicatorHitLog::getHitLadderValue).orElse(BigDecimal.ZERO) :
                        hitLog.map(EvalGroupIndicatorHitLog::getHitCommissionValue).orElse(BigDecimal.ZERO);
                Long evalUserReportId = shopPoolReportIdMap.get(pool.getId());
                reportUserValues.add(convertDB(evalUserReportId, groupIndicator, value, date, params));
            }

            //是否有子指标
            if (IndicatorCodeTypeEnum.COMBINE_INDICATOR.equals(codeType)) {
                for (EvalGroupIndicatorParam param : params) {
                    Map<Long, EvalGroupIndicatorTargetHitLog> targetHitLogMap = queryTargetValue(param.getId(), poolIds, date, EvalScopeEnum.SHOP);
                    for (EvalShopPool pool : pools) {
                        Boolean staffChange = pool.getDataDate().compareTo(date) != 0;
                        EvalGroupIndicatorTargetHitLog targetHitLog = (staffChange)
                                ? queryTargetValue(param.getId(), pool.getId(), pool.getDataDate(), EvalScopeEnum.SHOP) : targetHitLogMap.get(pool.getId());
                        BigDecimal value = this.getParamValue(targetHitLog);
                        Long evalUserReportId = shopPoolReportIdMap.get(pool.getId());
                        reportUserValues.add(convertParamDB(evalUserReportId, param, value, date));
                    }
                }
            }
        }
        if (PublicUtil.isNotEmpty(reportUserValues)) {
            evalUserValueReportService.saveBatch(reportUserValues);
        }
        return reportUserValues;
    }

    public BigDecimal getParamValue(EvalGroupIndicatorTargetHitLog hitLog) {
        if (PublicUtil.isEmpty(hitLog) || PublicUtil.isEmpty(hitLog.getUseTarget())) {
            return BigDecimal.ZERO;
        }
        BigDecimal value = null;
        switch (hitLog.getUseTarget()) {
            case NO: {
                value = Optional.ofNullable(hitLog.getValue()).orElse(BigDecimal.ZERO);
                break;
            }
            case FIRST_TARGET: {
                value = Optional.ofNullable(hitLog.getReachValue()).orElse(BigDecimal.ZERO);
                break;
            }
            case EXTRA_TARGET: {
                value = Optional.ofNullable(hitLog.getExtraReachValue()).orElse(BigDecimal.ZERO);
                break;
            }
        }

       return value;
    }

    public List<EvalUserReport> saveStaffReports(List<EvalUserPool> pools, LocalDate date, Long evalUserRankReportId, Long evalUserRankStageReportId) {
        List<EvalUserReport> evalUserReports = new ArrayList<>();
        for (EvalUserPool userPool : pools) {
            EvalUserReport report = new EvalUserReport();
            report.setEvalUserRankReportId(evalUserRankReportId);
            report.setEvalUserRankStageReportId(evalUserRankStageReportId);
            report.setDimension(ReportDimensionEnum.STAFF);
            report.setUserId(userPool.getUserId());
            report.setUserName(userPool.getUserName());
            report.setPostId(userPool.getPostId());
            report.setPostName(userPool.getPostName());
            report.setShopId(userPool.getShopId());
            report.setShopName(userPool.getShopName());
            report.setEvalUserPoolId(userPool.getId());
            report.setDataDate(date);
            report.setInclusion(userPool.getInclusion());
            evalUserReports.add(report);
        }
        evalUserReportService.saveBatch(evalUserReports);
        return evalUserReports;
    }

    public List<EvalUserReport> saveShopReports(List<EvalShopPool> shopPools, LocalDate date, Long evalUserRankReportId, Long evalUserRankStageReportId) {
        List<EvalUserReport> evalShopReports = new ArrayList<>();
        for (EvalShopPool shopPool : shopPools) {
            EvalUserReport report = new EvalUserReport();
            report.setDimension(ReportDimensionEnum.SHOP);
            report.setEvalUserRankReportId(evalUserRankReportId);
            report.setEvalUserRankStageReportId(evalUserRankStageReportId);
            report.setShopId(shopPool.getShopId());
            report.setShopName(shopPool.getShopName());
            report.setDataDate(date);
            evalShopReports.add(report);
        }
        evalUserReportService.saveBatch(evalShopReports);
        return evalShopReports;
    }

    public Map<Long, BigDecimal> queryUserOriginValue(EvalGroupIndicatorParam param, List<Long> userIds, LocalDate dataDate) {
        final String code = param.getCode();
        final IndicatorCodeTypeEnum codeType = param.getCodeType();
        Map<Long, BigDecimal> referValueMap = Maps.newHashMapWithExpectedSize(userIds.size());
        //无目标
        if (IndicatorCodeTypeEnum.INDICATOR.equals(codeType)) {
            List<IndicatorUserValue> indicatorUserValues = indicatorUserValueService.list(Wrappers.<IndicatorUserValue>lambdaQuery()
                            .eq(IndicatorUserValue::getIndicatorCode, code)
                            .eq(IndicatorUserValue::getDataDate, dataDate)
                            .in(IndicatorUserValue::getUserId, userIds)
                            .eq(IndicatorUserValue::getDimensionType, DimensionTypeEnum.STAFF)
                            .eq(IndicatorUserValue::getYn, Boolean.TRUE)
            );
            for (IndicatorUserValue userValue : indicatorUserValues) {
                referValueMap.put(userValue.getUserId(), commonService.queryIndicatorValueVO(code, userValue,
                        param.getTargetType(), param.getDataType()));
            }
            return referValueMap;
        }
        for (Long userId : userIds) {
            EvalIndicatorValue evalIndicatorValue = evalIndicatorValueService.queryLastValue(userId,
                    DimensionTypeEnum.STAFF, code, dataDate);
            referValueMap.put(userId, commonService.queryEvalIndicatorValueVO(code, evalIndicatorValue, param.getTargetType(), param.getDataType()));
        }
        return referValueMap;
    }

    public Map<Long, EvalGroupIndicatorTargetHitLog> queryTargetValue(Long referId, List<Long> poolIds, LocalDate dataDate, EvalScopeEnum scopeType) {
        List<EvalGroupIndicatorTargetHitLog> targetHitLogs = evalGroupIndicatorTargetHitLogService.list(Wrappers.<EvalGroupIndicatorTargetHitLog>lambdaQuery()
                .eq(EvalGroupIndicatorTargetHitLog::getReferId, referId)
                .eq(EvalGroupIndicatorTargetHitLog::getTargetType, IndicatorTypeEnum.EXAMINE)
                .eq(EvalGroupIndicatorTargetHitLog::getDataDate, dataDate)
                .in(EvalGroupIndicatorTargetHitLog::getPoolId, poolIds)
                .eq(EvalGroupIndicatorTargetHitLog::getScopeType, scopeType)
                .eq(EvalGroupIndicatorTargetHitLog::getYn, Boolean.TRUE)
        );
        return targetHitLogs.stream()
                .collect(Collectors.toMap(EvalGroupIndicatorTargetHitLog::getPoolId, Function.identity(), (v1, v2) -> v1));
    }

    public EvalGroupIndicatorTargetHitLog queryTargetValue(Long referId, Long poolId, LocalDate dataDate, EvalScopeEnum scopeType) {
        EvalGroupIndicatorTargetHitLog targetHitLog = evalGroupIndicatorTargetHitLogService.getOne(Wrappers.<EvalGroupIndicatorTargetHitLog>lambdaQuery()
                .eq(EvalGroupIndicatorTargetHitLog::getReferId, referId)
                .eq(EvalGroupIndicatorTargetHitLog::getTargetType, IndicatorTypeEnum.EXAMINE)
                .eq(EvalGroupIndicatorTargetHitLog::getDataDate, dataDate)
                .eq(EvalGroupIndicatorTargetHitLog::getPoolId, poolId)
                .eq(EvalGroupIndicatorTargetHitLog::getScopeType, scopeType)
                .eq(EvalGroupIndicatorTargetHitLog::getYn, Boolean.TRUE)
                , Boolean.FALSE
        );
        return targetHitLog;
    }

    public void cleanTodayData(LocalDate date) {
        evalUserRankReportService.remove(Wrappers.<EvalUserRankReport>lambdaUpdate()
                .eq(EvalUserRankReport::getDataDate, date)
        );
        evalUserRankCodeReportService.remove(Wrappers.<EvalUserRankCodeReport>lambdaUpdate()
                .eq(EvalUserRankCodeReport::getDataDate, date)
        );
        evalUserValueReportService.remove(Wrappers.<EvalUserValueReport>lambdaUpdate()
                .eq(EvalUserValueReport::getDataDate, date)
        );
        evalUserReportService.remove(Wrappers.<EvalUserReport>lambdaUpdate()
                .eq(EvalUserReport::getDataDate, date)
        );
    }

    public void cleanRankCurMonthData(Long rankId, LocalDate endDate) {
        LocalDate startDate = endDate.with(TemporalAdjusters.firstDayOfMonth());
        List<EvalUserRankReport> rankReports = evalUserRankReportService.list(Wrappers.<EvalUserRankReport>lambdaQuery()
                .eq(EvalUserRankReport::getRankId, rankId)
                .ge(EvalUserRankReport::getDataDate, startDate)
                .le(EvalUserRankReport::getDataDate, endDate)
        );
        if (PublicUtil.isEmpty(rankReports)) {
            return;
        }
        List<Long> rankReportIds = rankReports.stream().map(EvalUserRankReport::getId).collect(Collectors.toList());
        evalUserRankReportService.removeByIds(rankReportIds);
        evalUserRankCodeReportService.remove(Wrappers.<EvalUserRankCodeReport>lambdaUpdate()
                .in(EvalUserRankCodeReport::getEvalUserRankReportId, rankReportIds)
        );

        List<EvalUserReport> userReports = evalUserReportService.list(Wrappers.<EvalUserReport>lambdaQuery()
                .in(EvalUserReport::getEvalUserRankReportId, rankReportIds)
        );
        List<Long> userReportIds = userReports.stream().map(EvalUserReport::getId).collect(Collectors.toList());
        if (PublicUtil.isEmpty(userReportIds)) {
            return;
        }
        evalUserReportService.removeByIds(userReportIds);
        evalUserValueReportService.remove(Wrappers.<EvalUserValueReport>lambdaUpdate()
                .in(EvalUserValueReport::getEvalUserReportId, userReportIds)
        );
    }

    public List<EvalUserRankCodeReport> saveRankCodeReports(List<EvalGroupIndicator> evalGroupIndicators, Long reportRankId,
                                                            Long evalUserRankStageReportId, LocalDate date) {
        Set<String> codes = new HashSet<>();
        List<EvalUserRankCodeReport> rankCodes = new ArrayList<>();
        AtomicInteger order = new AtomicInteger();
        for (EvalGroupIndicator groupIndicator : evalGroupIndicators) {
            //一个编码只能展示一次
            if (! codes.add(groupIndicator.getCode())) {
                continue;
            }
            rankCodes.add(this.convertRankCodeDB(groupIndicator, date, reportRankId, evalUserRankStageReportId, order));

            //组合指标
            if (PublicUtil.isNotEmpty(groupIndicator.getCodeType()) &&
                    IndicatorCodeTypeEnum.COMBINE_INDICATOR.equals(groupIndicator.getCodeType())) {
                //子指标
                List<EvalGroupIndicatorParam> params = evalGroupIndicatorParamService.list(Wrappers.<EvalGroupIndicatorParam>lambdaQuery()
                        .eq(EvalGroupIndicatorParam::getEvalGroupIndicatorId, groupIndicator.getId())
                        .eq(EvalGroupIndicatorParam::getYn, Boolean.TRUE)
                );
                for (EvalGroupIndicatorParam indicatorParam : params) {
                    rankCodes.add(this.convertParamToRankCodeDB(indicatorParam, date, reportRankId, order, groupIndicator));
                }
            }

        }
        evalUserRankCodeReportService.saveBatch(rankCodes);
        return rankCodes;
    }

    public Map<Long, List<EvalGroupRank>> queryEffectRank(LocalDate date) {
        List<SettingStatusEnum> status = new ArrayList<SettingStatusEnum>(){{
            add(SettingStatusEnum.EFFECTIVE);
            add(SettingStatusEnum.BE_EFFECTIVE);
        }};
        List<EvalGroupRank> evalGroupRanks = evalGroupRankService.list(Wrappers.<EvalGroupRank>lambdaQuery()
                .in(EvalGroupRank::getStatus, status)
                .ge(EvalGroupRank::getOverTime, date)
                .eq(EvalGroupRank::getYn, Boolean.TRUE)
        );
        return evalGroupRanks.stream().collect(Collectors.groupingBy(EvalGroupRank::getGroupId));
    }

    public EvalUserRankReport saveRankReport(EvalGroupRank rank, LocalDate date, Long groupId) {
//        List<Long> shopIds = evalGroups.stream().map(EvalGroup::getShopIds).distinct().collect(ArrayList::new, List::addAll, List::addAll);
        EvalUserRankReport reportRank = new EvalUserRankReport();
        reportRank.setRankId(rank.getId());
        reportRank.setRankName(rank.getName());
        reportRank.setMultiStage(rank.getMultiStage());
        reportRank.setDataDate(date);
        reportRank.setMonthly(YearMonth.from(date));
        reportRank.setGroupId(groupId);
        evalUserRankReportService.save(reportRank);
        return reportRank;
    }

    public EvalUserRankStageReport saveRankStageReport(EvalUserRankReport rankReport, EvalGroupRank rank,
                                                       EvalGroupRankStage rankStage, List<EvalGroup> evalGroups) {
        List<Long> shopIds = evalGroups.stream().map(EvalGroup::getShopIds).distinct().collect(ArrayList::new, List::addAll, List::addAll);
        List<Long> evalGroupIds = evalGroups.stream().map(EvalGroup::getId).collect(Collectors.toList());
        EvalUserRankStageReport reportRankStage = new EvalUserRankStageReport();
        reportRankStage.setEvalUserRankReportId(rankReport.getId());
        reportRankStage.setEvalGroupRankId(rank.getId());
        reportRankStage.setEvalGroupRankStageId(rankStage.getId());
        if (PublicUtil.isNotEmpty(rankStage.getName())) {
            reportRankStage.setStageName(rankStage.getName());
        }
        reportRankStage.setEvalGroupIds(evalGroupIds);
        reportRankStage.setBeginTime(rankStage.getBeginTime());
        reportRankStage.setOverTime(rankStage.getOverTime());
        evalUserRankStageReportService.save(reportRankStage);
        return reportRankStage;
    }

    public EvalUserValueReport convertDB(Long evalUserReportId, EvalGroupIndicator groupIndicator, BigDecimal value,
                                         LocalDate date, List<EvalGroupIndicatorParam> params) {
        EvalUserValueReport userValue = new EvalUserValueReport();
        userValue.setEvalUserReportId(evalUserReportId);
        userValue.setCode(groupIndicator.getCode());
        userValue.setCodeType(groupIndicator.getCodeType());
        userValue.setValue(Optional.ofNullable(value).orElse(BigDecimal.ZERO));
        userValue.setDataDate(date);
        if (IndicatorCodeTypeEnum.COMBINE_INDICATOR.equals(groupIndicator.getCodeType())) {
            userValue.setValue(userValue.getValue().multiply(Constant.ONE_HUNDRED));
        } else {
            Optional<EvalGroupIndicatorParam> paramOptional = params.stream().findFirst();
            if (paramOptional.isPresent()) {
                EvalGroupIndicatorParam param = paramOptional.get();
                if (DataTypeEnum.RATIO.equals(param.getDataType()) || (! TargetTypeEnum.NO.equals(param.getTargetType()))) {
                    userValue.setValue(userValue.getValue().multiply(Constant.ONE_HUNDRED));
                }
            }
        }
        return userValue;
    }

    public EvalUserValueReport convertParamDB(Long evalUserReportId, EvalGroupIndicatorParam param, BigDecimal value, LocalDate date) {
        EvalUserValueReport userValue = new EvalUserValueReport();
        userValue.setEvalUserReportId(evalUserReportId);
        userValue.setCode(param.getCode());
        userValue.setCodeType(param.getCodeType());
        userValue.setValue(Optional.ofNullable(value).orElse(BigDecimal.ZERO));
        userValue.setDataDate(date);
        return userValue;
    }

    public EvalUserRankCodeReport convertRankCodeDB(EvalGroupIndicator groupIndicator, LocalDate date, Long reportRankId,
                                                    Long evalUserRankStageReportId, AtomicInteger order) {
        EvalUserRankCodeReport rankCode = new EvalUserRankCodeReport();
        rankCode.setEvalUserRankReportId(reportRankId);
        rankCode.setEvalUserRankStageReportId(evalUserRankStageReportId);
        rankCode.setEvalGroupIndicatorId(groupIndicator.getId());
        rankCode.setCode(groupIndicator.getCode());
        rankCode.setCodeType(groupIndicator.getCodeType());
//        String indicatorName = (ScoreWayEnum.LADDER.equals(groupIndicator.getScoreWay())) ?
//                groupIndicator.getLadderParamAlias() : groupIndicator.getCommissionParamAlias();
//        if (PublicUtil.isEmpty(indicatorName)) {
//            indicatorName = groupIndicator.getName();
//        }
//        rankCode.setIndicatorName(indicatorName);
        rankCode.setIndicatorName(groupIndicator.getName());
        rankCode.setDataDate(date);
        rankCode.setOrderNum(order.incrementAndGet());
        return rankCode;
    }

    public EvalUserRankCodeReport convertParamToRankCodeDB(EvalGroupIndicatorParam indicatorParam, LocalDate date, Long reportRankId,
                                                           AtomicInteger order, EvalGroupIndicator groupIndicator) {
        EvalUserRankCodeReport rankCode = new EvalUserRankCodeReport();
        rankCode.setEvalUserRankReportId(reportRankId);
        rankCode.setEvalGroupIndicatorId(indicatorParam.getEvalGroupIndicatorId());
        rankCode.setCode(indicatorParam.getCode());
        rankCode.setCodeType(indicatorParam.getCodeType());
        rankCode.setIndicatorName(indicatorParam.getName());
        if (IndicatorCodeTypeEnum.COMBINE_INDICATOR.equals(groupIndicator.getCodeType())
                && PublicUtil.isNotEmpty(groupIndicator.getCode())) {
            rankCode.setParentCode(groupIndicator.getCode());
        }
        rankCode.setDataDate(date);
        rankCode.setOrderNum(order.incrementAndGet());
        return rankCode;
    }

}
